@@ -68,6 +68,7 @@ Format metadata has the following form:
 
 # Python imports
 from abc import ABC, abstractmethod
+from collections import defaultdict
 import copy
 import csv
 import io
@@ -187,6 +188,14 @@ class FormatRegistry:
             raise QuiltException("No object type or metadata specified, and guessing by extension failed.")
         return ext_fmts
 
+    @classmethod
+    def object_is_serializable(cls, obj):
+        try:
+            format_handlers = cls.search(type(obj))
+            return True
+        except QuiltException as e:
+            return False
+
     @classmethod
     def serialize(cls, obj, meta=None, ext=None, **format_opts):
         """Match an object to a format, and serialize it to that format.
@@ -304,6 +313,43 @@ class FormatRegistry:
         name = cls._get_name_from_meta(meta)
         return cls.for_format(name)
 
+    @classmethod
+    def all_supported_formats(cls):
+        """
+        Returns a map of supported object types -> serialization formats, e.g.:
+
+                            Python Object Type     Serialization Formats
+         <class 'pandas.core.frame.DataFrame'>  [ssv, csv, tsv, parquet]
+                       <class 'numpy.ndarray'>                [npy, npz]
+                                 <class 'str'>      [md, json, rst, txt]
+                                <class 'dict'>                    [json]
+                            <class 'NoneType'>                    [json]
+                               <class 'tuple'>                    [json]
+                                 <class 'int'>                    [json]
+                                <class 'list'>                    [json]
+                               <class 'float'>                    [json]
+                               <class 'bytes'>                     [bin]
+        """
+        try:
+            import numpy as np
+        except ImportError:
+            pass
+        else:
+            cls.search(np.ndarray)  # Force FormatHandlers to register np.ndarray as a supported object type
+
+        try:
+            import pandas as pd
+        except ImportError:
+            pass
+        else:
+            cls.search(pd.DataFrame)  # Force FormatHandlers to register pd.DataFrame as a supported object type
+
+        type_map = defaultdict(set)
+        for handler in cls.registered_handlers:
+            for t in handler.handled_types:
+                type_map[t].update(handler.handled_extensions)
+        return dict(type_map)
+
 
 class BaseFormatHandler(ABC):
     """Base class for binary format handlers
@@ -887,7 +933,10 @@ class NumpyFormatHandler(BaseFormatHandler):
         return buf.getvalue(), self._update_meta(meta)
 
     def deserialize(self, bytes_obj, meta=None, ext=None, **format_opts):
-        import numpy as np
+        try:
+            import numpy as np
+        except ImportError:
+            raise QuiltException("Please install numpy")
 
         # security
         kwargs = dict(allow_pickle=False)
@@ -918,7 +967,7 @@ class ParquetFormatHandler(BaseFormatHandler):
     handled_extensions = ['parquet']
     opts = ('compression',)
     defaults = {
-        'compression': 'snappy_columns',
+        'compression': 'snappy',
     }
 
     def handles_type(self, typ):
@@ -941,29 +990,18 @@ class ParquetFormatHandler(BaseFormatHandler):
         opts = self.get_opts(meta, format_opts)
         opts_with_defaults = copy.deepcopy(self.defaults)
         opts_with_defaults.update(opts)
-        kwargs = {}
         table = pa.Table.from_pandas(obj)
-
-        for name, value in opts_with_defaults.items():
-            if name == 'compression':
-                if isinstance(value, str) and value.endswith('_columns'):
-                    # shorthand for columnar compression on all columns, using prefix value.
-                    compression = value.split('_')[0]
-                    kwargs['compression'] = {
-                        col.name.encode('utf-8'): compression for col in table.columns
-                    }
-                else:
-                    kwargs['compression'] = value
-
         buf = io.BytesIO()
-        parquet.write_table(table, buf, **kwargs)
+        parquet.write_table(table, buf, **opts_with_defaults)
 
         return buf.getvalue(), self._update_meta(meta, additions=opts_with_defaults)
 
-
     def deserialize(self, bytes_obj, meta=None, ext=None, **format_opts):
-        import pyarrow as pa
-        from pyarrow import parquet
+        try:
+            import pyarrow as pa
+            from pyarrow import parquet
+        except ImportError:
+            raise QuiltException("Please install pyarrow")
 
         buf = io.BytesIO(bytes_obj)
         table = parquet.read_table(buf)
