@@ -1,3 +1,4 @@
+from collections import deque
 import copy
 import hashlib
 import io
@@ -5,8 +6,10 @@ import json
 import pathlib
 import os
 import time
+from multiprocessing import Pool
 from urllib.parse import quote, urlparse, unquote
-from collections import deque
+import uuid
+import warnings
 
 import jsonlines
 
@@ -18,8 +21,9 @@ from .exceptions import PackageException
 from .formats import FormatRegistry
 from .util import (
     QuiltException, fix_url, get_from_config, get_install_location, make_s3_url, parse_file_url,
-    parse_s3_url, validate_package_name, quiltignore_filter, validate_key
+    parse_s3_url, validate_package_name, quiltignore_filter, validate_key, extract_file_extension, file_is_local
 )
+from .util import TEMPFILE_DIR_PATH as APP_DIR_TEMPFILE_DIR
 
 
 def hash_file(readable_file):
@@ -54,6 +58,12 @@ def _to_singleton(physical_keys):
     return physical_keys[0]
 
 
+def _delete_local_physical_key(pk):
+    assert file_is_local(pk), "This function only works on files that live on a local disk"
+    pathlib.Path(parse_file_url(urlparse(pk))).unlink()
+
+
+
 class PackageEntry(object):
     """
     Represents an entry at a logical key inside a package.
@@ -622,6 +632,11 @@ class Package(object):
             for obj in objects:
                 if not obj['IsLatest']:
                     continue
+                # Skip S3 pseduo directory files and Keys that end in /
+                if obj['Key'].endswith('/'):
+                    if obj['Size'] != 0:
+                        warnings.warn(f'Logical keys cannot end in "/", skipping: {obj["Key"]}')
+                    continue
                 obj_url = make_s3_url(src_bucket, obj['Key'], obj.get('VersionId'))
                 entry = PackageEntry([obj_url], obj['Size'], None, None)
                 logical_key = obj['Key'][len(src_key):]
@@ -649,7 +664,7 @@ class Package(object):
         if logical_key:
             obj = self[logical_key]
             if not isinstance(obj, PackageEntry):
-                raise ValueError("Key does point to a PackageEntry")
+                raise ValueError("Key does not point to a PackageEntry")
             return obj.get()
         else:
             # a package has a logical root directory if all of its children are rooted at the
@@ -719,6 +734,8 @@ class Package(object):
 
             return hypothesized_root_path
 
+    # def get_as_pathlib(self):
+
     def set_meta(self, meta):
         """
         Sets user metadata on this Package.
@@ -826,7 +843,7 @@ class Package(object):
     @property
     def manifest(self):
         """
-        Provides a generator of the dicts that make up the serialied package.
+        Provides a generator of the dicts that make up the serialized package.
         """
         yield self._meta
         for dir_key, meta in self._walk_dir_meta():
@@ -834,16 +851,26 @@ class Package(object):
         for logical_key, entry in self.walk():
             yield {'logical_key': logical_key, **entry.as_dict()}
 
-    def set(self, logical_key, entry=None, meta=None):
+    def set(self, logical_key, entry=None, meta=None, serialization_location=None, serialization_format_opts=None):
         """
         Returns self with the object at logical_key set to entry.
 
         Args:
             logical_key(string): logical key to update
-            entry(PackageEntry OR string): new entry to place at logical_key in the package.
+            entry(PackageEntry OR string OR object): new entry to place at logical_key in the package.
                 If entry is a string, it is treated as a URL, and an entry is created based on it.
                 If entry is None, the logical key string will be substituted as the entry value.
+                If entry is an object and quilt knows how to serialize it, it will immediately be serialized and written
+                to disk, either to serialization_location or to a location managed by quilt. List of types that Quilt
+                can serialize is available by calling `quilt3.formats.FormatRegistry.all_supported_formats()`
             meta(dict): user level metadata dict to attach to entry
+            serialization_format_opts(dict): Optional. If passed in, only used if entry is an object. Options to help
+                Quilt understand how the object should be serialized. Useful for underspecified file formats like csv
+                when content contains confusing characters. Will be passed as kwargs to the FormatHandler.serialize()
+                function. See docstrings for individual FormatHandlers for full list of options -
+                https://github.com/quiltdata/quilt/blob/master/api/python/quilt3/formats.py
+            serialization_location(string): Optional. If passed in, only used if entry is an object. Where the
+                serialized object should be written, e.g. "./mydataframe.parquet"
 
         Returns:
             self
@@ -856,7 +883,7 @@ class Package(object):
 
         validate_key(logical_key)
 
-        if not entry:
+        if entry is None:
             current_working_dir = pathlib.Path.cwd()
             logical_key_abs_path = pathlib.Path(logical_key).absolute()
             entry = logical_key_abs_path.relative_to(current_working_dir)
@@ -865,7 +892,7 @@ class Package(object):
             url = fix_url(str(entry))
             size, orig_meta, version = get_size_and_meta(url)
 
-            # Deterimine if a new version needs to be appended.
+            # Determine if a new version needs to be appended.
             parsed_url = urlparse(url)
             if parsed_url.scheme == 's3':
                 bucket, key, current_version = parse_s3_url(parsed_url)
@@ -874,10 +901,60 @@ class Package(object):
             entry = PackageEntry([url], size, None, orig_meta)
         elif isinstance(entry, PackageEntry):
             entry = entry._clone()
+
+        elif FormatRegistry.object_is_serializable(entry):
+            # Use file extension from serialization_location, fall back to file extension from logical_key
+            # If neither has a file extension, Quilt picks the serialization format.
+            logical_key_ext = extract_file_extension(logical_key)
+
+            serialize_loc_ext = None
+            if serialization_location is not None:
+                serialize_loc_ext = extract_file_extension(serialization_location)
+
+            if logical_key_ext is not None and serialize_loc_ext is not None:
+                assert logical_key_ext == serialize_loc_ext, f"The logical_key and the serialization_location have " \
+                                                             f"different file extensions: {logical_key_ext} vs " \
+                                                             f"{serialize_loc_ext}. Quilt doesn't know which to use!"
+
+            if serialize_loc_ext is not None:
+                ext = serialize_loc_ext
+            elif logical_key_ext is not None:
+                ext = logical_key_ext
+            else:
+                ext = None
+
+            format_handlers = FormatRegistry.search(type(entry))
+            if ext:
+                format_handlers = [f for f in format_handlers if ext in f.handled_extensions]
+
+            if len(format_handlers) == 0:
+                error_message = f'Quilt does not know how to serialize a {type(entry)}'
+                if ext is not None:
+                    error_message += f' as a {ext} file.'
+                error_message += f'. If you think this should be supported, please open an issue or PR at ' \
+                                 f'https://github.com/quiltdata/quilt'
+                raise QuiltException(error_message)
+
+            if serialization_format_opts is None:
+                serialization_format_opts = {}
+            serialized_object_bytes, new_meta = format_handlers[0].serialize(entry, meta=None, ext=ext,
+                                                                             **serialization_format_opts)
+            if serialization_location is None:
+                serialization_path = APP_DIR_TEMPFILE_DIR / str(uuid.uuid4())
+                if ext:
+                    serialization_path = serialization_path.with_suffix(f'.{ext}')
+            else:
+                serialization_path = pathlib.Path(serialization_location).expanduser().resolve()
+
+            serialization_path.parent.mkdir(exist_ok=True, parents=True)
+            serialization_path.write_bytes(serialized_object_bytes)
+
+            size = serialization_path.stat().st_size
+            write_url = serialization_path.as_uri()
+            entry = PackageEntry([write_url], size, hash_obj=None, meta=new_meta)
+
         else:
-            raise TypeError(
-                f"Expected a string for entry, but got an instance of {type(entry)}."
-            )
+            raise TypeError(f"Expected a string for entry, but got an instance of {type(entry)}.")
 
         if meta is not None:
             entry.set_meta(meta)
@@ -1022,12 +1099,29 @@ class Package(object):
 
         self._fix_sha256()
         pkg = self._materialize(dest)
+
+        def physical_key_is_temp_file(pk):
+            if not file_is_local(pk):
+                return False
+            return pathlib.Path(parse_file_url(urlparse(pk))).parent == APP_DIR_TEMPFILE_DIR
+
+        temp_file_logical_keys = [lk for lk, entry in self.walk() if physical_key_is_temp_file(entry.physical_keys[0])]
+        temp_file_physical_keys = [self.get(lk) for lk in temp_file_logical_keys]
+
+        # Now that data has been pushed, delete tmp files created by pkg.set('KEY', obj)
+        with Pool(10) as p:
+            p.map(_delete_local_physical_key, temp_file_physical_keys)
+
+        # Update old package to point to the materialized location of the file since the tempfile no longest exists
+        for lk in temp_file_logical_keys:
+            self.set(lk, pkg[lk])
+
         pkg.build(name, registry=registry, message=message)
         return pkg
 
     def _materialize(self, dest_url):
         """
-        Copies objects to path, then creates a new package that points to those objects.
+        Copies all Package entries to the destination, then creates a new package that points to those objects.
 
         Copies each object in this package to path according to logical key structure,
         and returns a package with physical_keys that point to the new copies.
@@ -1047,22 +1141,30 @@ class Package(object):
         pkg._meta = self._meta
         # Since all that is modified is physical keys, pkg will have the same top hash
         file_list = []
+        entries = []
         for logical_key, entry in self.walk():
             # Copy the datafiles in the package.
             physical_key = _to_singleton(entry.physical_keys)
+            unversioned_physical_key = physical_key.split('?', 1)[0]
             new_physical_key = dest_url + "/" + quote(logical_key)
-            file_list.append((physical_key, new_physical_key, entry.size, entry.meta))
+            new_entry = entry._clone()
+            if unversioned_physical_key == new_physical_key:
+                # No need to copy - re-use the original physical key.
+                pkg.set(logical_key, new_entry)
+            else:
+                entries.append((logical_key, new_entry))
+                file_list.append((physical_key, new_physical_key, entry.size, None))
 
         results = copy_file_list(file_list)
 
-        for (logical_key, entry), versioned_key in zip(self.walk(), results):
+        for (logical_key, new_entry), versioned_key in zip(entries, results):
             # Create a new package entry pointing to the new remote key.
             assert versioned_key is not None
-            new_entry = entry._clone()
             new_entry.physical_keys = [versioned_key]
             pkg.set(logical_key, new_entry)
         return pkg
 
+
     def diff(self, other_pkg):
         """
         Returns three lists -- added, modified, deleted.
@@ -1096,7 +1198,7 @@ class Package(object):
         Performs a user-specified operation on each entry in the package.
 
         Args:
-            f: function
+            f(x, y): function
                 The function to be applied to each package entry.
                 It should take two inputs, a logical key and a PackageEntry.
             include_directories: bool
@@ -1118,8 +1220,9 @@ class Package(object):
         removing results that evaluate to False from the output.
 
         Args:
-            f: function
+            f(x, y): function
                 The function to be applied to each package entry.
+                It should take two inputs, a logical key and a PackageEntry.
                 This function should return a boolean.
             include_directories: bool
                 Whether or not to include directory entries in the map.
